import {combineReducers,createStore} from 'redux'
import { shallow, mount, render } from 'enzyme';

describe('combineReducers',function(){
	it('returns a composite reducer that maps the state keys to given reducers',function(){

		const reducer = combineReducers({
			counter: (state = 0, action) =>
			action.type ==='increment' ? state + 1: state,
			stack: (state= [],action) =>
			action.type ==='push' ? [...state,action.value] : state
		})

		const a1 = reducer({},{type:'increment'})
		expect(a1).toEqual({counter :1,stack : []});

		const a2 = reducer(a1,{type : 'push',value : 'a'})
		expect(a2).toEqual({
			counter : 1,
			stack : ['a']
		});
	})

	it('ignores all props which are not a function',function(){
		const reducer = combineReducers({
			fake : true,
			broke : 'string',
			anthor : {method :'get'},
			stack : (state=[]) =>state
		})

		expect(Object.keys(reducer({},{type : 'push'}))).toEqual(['stack'])
	})

	it('throws an error if a reducer returns undefined handling an action',function(){
		const reducer = combineReducers({
			counter(state=0,action){
				switch(action && action.type){
					case 'increment' : 
						return state + 1
					case 'decrement' : 
						return state -1
					case 'whatever' : 
					//case 'null':
					case undefined : 
						return undefined;
					default : 
						return state
				}
			}
		})
		expect(()=>reducer({counter : 0},{type : 'whatever'})).toThrow();
		//expect(()=>reducer({counter :0},null)).toThrow()
		expect(()=>reducer({counter:0},{})).toThrow()
	})

	it('throws an error on first call if a reducer returns undefined initializing',function(){
		const reducer = combineReducers({
			counter(state,action){
				switch(action.type){
					case 'increment' : 
						return state +1
					case 'decrement' : 
						return state -1
					default : 
						return state
				}
			}
		})

		expect(()=>reducer({})).toThrow()
	})

	it('catches error thrown in reducer initializing and re-throw',function(){
		const reducer = combineReducers({
			throwingReducer(){
				throw new Error('Error thrown in reducer')
			}
		})

		expect(()=>reducer({})).toThrow();
	})


	it('allows a symbol to be used as an action type',function(){
		const increment = Symbol('INCREMENT')
		const reducer = combineReducers({
			counter(state=0,action){
				switch(action.type){
					case increment : 
						return state + 1
					default : 
						return state
				}
			}
		})

		expect(reducer({counter : 0},{type : increment}).counter).toEqual(1)
	})

	it('maintains referential equality if the reducers it is combining do',function(){
		const reducer = combineReducers({
			child1(state = {}){
				return state
			},
			child2(state ={}){
				return state
			},
			child3(state ={}){
				return state
			}
		})

		const initialState = reducer(undefined,'@@INIT')

		expect(reducer(initialState,{type:'Foo'})).toEqual(initialState)
	})

	it('warns if input state does not match reducer shape',function(){
		const spy = spyOn(console,'error')
		const reducer = combineReducers({
			foo(state={bar:1}){
				return state
			},
			baz(state ={type : 3}){
				return state
			}
		})
		
		reducer()
		expect(spy.calls.count()).toBe(0)
		
		reducer({foo:{bar:2}})
		expect(spy.calls.count()).toBe(0)
		
		reducer({baz : {type :2}})
		expect(spy.calls.count()).toBe(0)

		reducer({
			foo : {bar:1},
			baz : {type :1}
		})
		expect(spy.calls.count()).toBe(0)

		createStore(reducer,{bar : 2})

		expect(spy.calls.allArgs()[0][0]).toMatch( /Unexpected key "bar".*createStore.*instead: "foo", "baz"/);

		createStore(reducer, { bar: 2, qux: 4 })

		
      	expect(spy.calls.allArgs()[0][0]).toMatch(/Unexpected key "bar" found in initialState argument passed to createStore. Expected to find one of the known reducer keys instead: "foo", "baz". Unexpected keys will be ignored./)

      	createStore(reducer, 1)
      	expect(spy.calls.allArgs()[0][0]).toMatch(/Unexpected key "bar" found in initialState argument passed to createStore. Expected to find one of the known reducer keys instead: "foo", "baz". Unexpected keys will be ignored./)

      	reducer(1)

      	expect(spy.calls.allArgs()[0][0]).toMatch(/Unexpected key "bar" found in initialState argument passed to createStore. Expected to find one of the known reducer keys instead: "foo", "baz". Unexpected keys will be ignored./)


	})
});