import { createStore, combineReducers} from 'redux'
import {todos,addToDo,errorThrowingReducer,dispatchInMiddle,dispatchInTheMiddleOfReducer,throwError,unknowAction,todosReverse} from '../testRely/reducer.js'
import { shallow, mount, render } from 'enzyme';

describe('createStore',function(){
	it('exposes the public api',function(){

		const store = createStore(combineReducers({todos}))
		const methods = Object.keys(store)

		expect(methods.length).toBe(4)
		expect(methods).toContain('subscribe')
		expect(methods).toContain('dispatch')
		expect(methods).toContain('getState')
		expect(methods).toContain('replaceReducer')
	})

	it('throws if reducer is not a function',function(){
		
		expect(()=>createStore()).toThrow()
		expect(()=>createStore('test')).toThrow()
		expect(()=>createStore({})).toThrow()
		expect(()=>createStore(()=>{})).not.toThrow()
	})

	it('passes the initial action and the initial state',function(){
		const store = createStore(todos,[
		{
			id : 1,
			text : 'one'
		}
		]);

		expect(store.getState()).toEqual([{
			id : 1,
			text : 'one'
		}]);
	})

	it('applies th reducer to the previous state',function(){
		const store = createStore(todos)
		expect(store.getState()).toEqual([])

		store.dispatch(addToDo('one'))
		expect(store.getState()).toEqual([{
			id : 1,
			text :"one"
		}])

		store.dispatch(unknowAction())
		expect(store.getState()).toEqual([{
			id : 1,
			text :"one"
		}])

		store.dispatch(addToDo('two'))
		expect(store.getState()).toEqual([{
			id : 1,
			text : 'one'
		},{
			id : 2,
			text : 'two'
		}])

	})

	it('preserves the state when replacing a reducer',function(){
		const store = createStore(todos)
		store.dispatch(addToDo('one'))
		expect(store.getState()).toEqual([{
			id : 1,
			text : 'one'
		}])

		store.replaceReducer(todosReverse)
		store.dispatch(addToDo('two'))
		expect(store.getState()).toEqual([{
			id : 2,
			text : 'two'
		},{
			id : 1,
			text : 'one'
		}])

		store.replaceReducer(todos)
		expect(store.getState()).toEqual([{
			id : 2,
			text : 'two'
		},{
			id : 1,
			text : 'one'
		}])
		store.dispatch(addToDo('three'))		
		expect(store.getState()).toEqual([{
			id : 2,
			text : 'two'
		},{
			id : 1,
			text : 'one'
		},{
			id : 3,
			text : 'three'
		}])
	})

	it('supports multiple subscriptions',function(){
		const store = createStore(todos)
		const listenA = jasmine.createSpy()
		const listenB = jasmine.createSpy()
		let unsubscribeA = store.subscribe(listenA)

		store.dispatch(unknowAction())
		expect(listenA.calls.count()).toBe(1)
		expect(listenB.calls.count()).toBe(0)

		store.dispatch(unknowAction())
		expect(listenA.calls.count()).toBe(2)
		expect(listenB.calls.count()).toBe(0)

		const unsubscribeB = store.subscribe(listenB)

		expect(listenB.calls.count()).toBe(0)
		expect(listenA.calls.count()).toBe(2)

		store.dispatch(unknowAction())
		expect(listenA.calls.count()).toBe(3)
		expect(listenB.calls.count()).toBe(1)

		unsubscribeA()

		expect(listenA.calls.count()).toBe(3)
		expect(listenB.calls.count()).toBe(1)

		store.dispatch(unknowAction())
    	expect(listenA.calls.count()).toBe(3)
    	expect(listenB.calls.count()).toBe(2)

		unsubscribeB()
		store.dispatch(unknowAction())
		expect(listenA.calls.count()).toBe(3)
		expect(listenB.calls.count()).toBe(2)

		unsubscribeA = store.subscribe(listenA)
		expect(listenA.calls.count()).toBe(3)
		expect(listenB.calls.count()).toBe(2)

		store.dispatch(unknowAction())
		expect(listenA.calls.count()).toBe(4)
		expect(listenB.calls.count()).toBe(2)

	})
	
	it('only removes listener once when unsubscribe is called',function(){
		const store = createStore(todos)
		const listenA = jasmine.createSpy()
		const listenB = jasmine.createSpy()

		const unsubscribeA = store.subscribe(listenA)
		store.subscribe(listenB)

		unsubscribeA()

		store.dispatch(unknowAction())
		expect(listenA.calls.count()).toBe(0)
		expect(listenB.calls.count()).toBe(1)
	})

	it('only removes relevant listener when unsubscribe is called',function(){
		const store = createStore(todos)
		const listener = jasmine.createSpy()
		
		store.subscribe(listener)

		const unsubscribeSecond = store.subscribe(listener)

		unsubscribeSecond()

		store.dispatch(unknowAction())
		expect(listener.calls.count()).toBe(1)
	})

	it('supports removing a subscription within a subscription',function(){

		const store = createStore(todos)
		const listenA = jasmine.createSpy()
		const listenB = jasmine.createSpy()
		const listenC = jasmine.createSpy()

		store.subscribe(listenA)
		const unSubB = store.subscribe(()=>{
			listenB(),
			unSubB()
		})

		store.subscribe(listenC)
		store.dispatch(unknowAction())

		expect(listenA.calls.count()).toBe(1)
		expect(listenB.calls.count()).toBe(1)
		expect(listenC.calls.count()).toBe(1)

	})
	it('delays unsubscribe until the end of current dispatch',function(){
		const store = createStore(todos)
		const unsubscribeHandles = []
		const doUnsubscribeAll = ()=>unsubscribeHandles.forEach(
			unsubscribe =>unsubscribe()
		)

		const listenA = jasmine.createSpy()
		const listenB = jasmine.createSpy()
		const listenC = jasmine.createSpy()

		unsubscribeHandles.push(store.subscribe(()=>listenA()))
		unsubscribeHandles.push(store.subscribe(()=>{
			listenB()
			doUnsubscribeAll()
		}))
		unsubscribeHandles.push(store.subscribe(()=>listenC()))

		store.dispatch(unknowAction())
		expect(listenA.calls.count()).toBe(1)
		expect(listenB.calls.count()).toBe(1)
		expect(listenC.calls.count()).toBe(1)

		store.dispatch(unknowAction())

		expect(listenA.calls.count()).toBe(1)
		expect(listenB.calls.count()).toBe(1)
		expect(listenC.calls.count()).toBe(1)
	})

	it('delays subscribe until the end of current dispatch',function(){
		const store = createStore(todos)
		const listenA = jasmine.createSpy()
		const listenB = jasmine.createSpy()
		const listenC = jasmine.createSpy()

		let listenCAdded = false
		const maybeAddThirdListener = ()=>{
			if(!listenCAdded){
				listenCAdded = true
				store.subscribe(()=>listenC())
			}
		}

		store.subscribe(()=>listenA())
		store.subscribe(()=>{
			listenB()
			maybeAddThirdListener()
		})

		store.dispatch(unknowAction())
		expect(listenA.calls.count()).toBe(1)
		expect(listenB.calls.count()).toBe(1)
		expect(listenC.calls.count()).toBe(0)

		store.dispatch(unknowAction())
		expect(listenA.calls.count()).toBe(2)
		expect(listenB.calls.count()).toBe(2)
		expect(listenC.calls.count()).toBe(1)
	})

	it('uses the last snapshot of subsucribers during nested dispatch',function(){
		const store = createStore(todos)
		const listenA = jasmine.createSpy()
		const listenB = jasmine.createSpy()
		const listenC = jasmine.createSpy()
		const listenD = jasmine.createSpy()

		let unsubscribe4;
		const unsubscribe1 = store.subscribe(()=>{
			listenA()
			expect(listenA.calls.count()).toBe(1)
			expect(listenB.calls.count()).toBe(0)
			expect(listenC.calls.count()).toBe(0)
			expect(listenD.calls.count()).toBe(0)

			unsubscribe1()

			unsubscribe4 = store.subscribe(listenD)
			store.dispatch(unknowAction())
			expect(listenA.calls.count()).toBe(1)
			expect(listenB.calls.count()).toBe(1)
			expect(listenC.calls.count()).toBe(1)
			expect(listenD.calls.count()).toBe(1)
		})
		
		store.subscribe(listenB)
		store.subscribe(listenC)
		store.dispatch(unknowAction())
		expect(listenA.calls.count()).toBe(1)
		expect(listenB.calls.count()).toBe(2)
		expect(listenC.calls.count()).toBe(2)
		expect(listenD.calls.count()).toBe(1)

		unsubscribe4()
		store.dispatch(unknowAction())
		expect(listenA.calls.count()).toBe(1)
		expect(listenB.calls.count()).toBe(3)
		expect(listenC.calls.count()).toBe(3)
		expect(listenD.calls.count()).toBe(1)
	})

	it('provides an up-to-date state when a subscriber is notified',function(){
		const store = createStore(todos)
		expect(()=>store.dispatch(unknowAction())).not.toThrow()
		function awesSomeMap(){
			[null,undefined,42,'hey',new awesSomeMap()].forEach(nonObject=>expect(()=>store.dispatch(nonObject).toThrow(/plain/)))
		}
	})

	it('handles nested dispatches gracefully',function(){
		function foo(state=0,action){
			return action.type ==='foo' ? 1: state
		}
		function bar(state=0,action){
			return action.type ==='bar' ? 2: state
		}
		const store = createStore(combineReducers({foo,bar}))
		store.subscribe(function kindaComponentDidUpdate(){
			const state = store.getState()
			if(state.bar ===0){
				store.dispatch({type : 'bar'})
			}
		})
		store.dispatch({type : 'foo'})
		expect(store.getState()).toEqual({
			foo : 1,
			bar : 2
		})
	})

	it('does not allow dispatch() form within a reducer',function(){
		const store = createStore(dispatchInTheMiddleOfReducer)
		expect(()=>{
			store.dispatch(dispatchInMiddle(store.dispatch.bind(store,unknowAction())))
		}).toThrow()
	})

	it('recovers from an error within a reducer',function(){
		const store = createStore(errorThrowingReducer)
		expect(()=>{
			store.dispatch(throwError()).toThrow()
		})
		expect(()=>{
			store.dispatch(unknowAction()).not.toThrow()
		})
	})

	it('throws if action type is missing',function(){
		const store = createStore(todos)
		expect(()=>{
			store.dispatch({}).toThrow()
		})
	})

	it('throws if action type is undefined',function(){
		const store = createStore(todos)
		expect(()=>{
			store.dispatch({type:'undefined'}).toThrow()
		})
	})

	it('does not throw if action type is falsy',function(){
		const store = createStore(todos)
		expect(()=>{
			store.dispatch({type:false}).not.toThrow()
		})
		expect(()=>{
			store.dispatch({type :0}).not.toThrow()
		})
		expect(()=>{
			store.dispatch({type:null}).not.toThrow()
		})
		expect(()=>{
			store.dispatch({type:''}).not.toThrow()
		})
	})
})